package org.svaor.preferences;

import org.apache.commons.lang.ArrayUtils;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.cache.Cache;
import org.springframework.cache.CacheManager;

import java.util.prefs.BackingStoreException;

import static org.fluentjava.FluentUtils.cast;

/**
 * Caching PreferencesProvider for remote client (to decrease remote server calls count)
 * @author Smelik Evgeniy
 */
public class ClientCachingPreferencesServiceImpl implements PreferencesService, InitializingBean {
	private static final String CACHE_NAME_ROOT_NODE = "pref_root_node";
	private static final String CACHE_NAME_NODE = "pref_node";
	private static final String CACHE_NAME_NODE_NAMES = "pref_node_names";
	private static final String CACHE_NAME_PROPERTY = "pref_property";
	private static final String CACHE_NAME_PROPERTY_KEYS = "pref_property_keys";

	private static final String CACHE_KEY_SEPARATOR = "\n";

	//spring
	private PreferencesService remotePreferencesService;
	private CacheManager cacheManager;

	//caches
	private Cache cacheRootNode;
	private Cache cacheNode;
	private Cache cacheNodeNames;
	private Cache cacheProperty;
	private Cache cachePropertyKeys;

	public ClientCachingPreferencesServiceImpl() {
	}

	@NotNull
	@Override
	public Long loadOrCreateRootNode(@Nullable final String username) {
		class CallbackLoadOrCreateNode implements Callback<Long, RuntimeException> {
			@Override
			public Long call() {
				return remotePreferencesService.loadOrCreateRootNode(username);
			}
		}

		final CallbackLoadOrCreateNode callback = new CallbackLoadOrCreateNode();
		return getValue(callback, cacheRootNode, username);
	}

	@NotNull
	@Override
	public Long[] setProperty(@NotNull Long rootNodeId, @Nullable String[] nodeNames, @NotNull String nodePath, @NotNull String key, @NotNull String value) {
		final Long[] createdIds = remotePreferencesService.setProperty(rootNodeId, nodeNames, nodePath, key, value);
		putValue(cacheProperty, nodePath + CACHE_KEY_SEPARATOR + key, value);
		evictValue(cachePropertyKeys, nodePath);
		return createdIds;
	}

	@Nullable
	@Override
	public String getProperty(@NotNull final String nodePath, @Nullable final Long nodeId, @NotNull final String key) {
		class CallbackGetProperty implements Callback<String, RuntimeException> {
			@Nullable
			@Override
			public String call() {
				return remotePreferencesService.getProperty(nodePath, nodeId, key);
			}
		}

		final CallbackGetProperty callback = new CallbackGetProperty();
		return getValue(callback, cacheProperty, nodePath + CACHE_KEY_SEPARATOR + key);
	}

	@Override
	public void removeProperty(@NotNull final String nodePath, @NotNull Long nodeId, @NotNull String key) {
		remotePreferencesService.removeProperty(nodePath, nodeId, key);
		evictValue(cacheProperty, nodePath + CACHE_KEY_SEPARATOR + key);
		evictValue(cachePropertyKeys, nodePath);
	}

	@Override
	public void removeNode(@NotNull String parentNodePath, @NotNull String nodeName, @NotNull Long nodeId) throws BackingStoreException {
		remotePreferencesService.removeNode(parentNodePath, nodeName, nodeId);

		evictValue(cacheNode, parentNodePath + CACHE_KEY_SEPARATOR + nodeName);

		final String nodePath = parentNodePath + PREFERENCES_NAME_SEPARATOR + nodeName;
		evictValue(cacheNodeNames, nodePath);

		final String[] propertyKeys = getValue(new EmptyCallback<String[], BackingStoreException>(), cachePropertyKeys, nodePath);
		if (ArrayUtils.isNotEmpty(propertyKeys)) {
			for (String propertyKey : propertyKeys) {
				evictValue(cacheProperty, nodePath + CACHE_KEY_SEPARATOR + propertyKey);
			}
		}

		evictValue(cachePropertyKeys, nodePath);
	}

	@NotNull
	@Override
	public String[] getPropertyKeys(@NotNull final String nodePath, @Nullable final Long nodeId) throws BackingStoreException {
		class CallbackPropertyKeys implements Callback<String[], BackingStoreException> {
			@Override
			public String[] call() throws BackingStoreException {
				return remotePreferencesService.getPropertyKeys(nodePath, nodeId);
			}
		}

		final CallbackPropertyKeys callback = new CallbackPropertyKeys();
		return getValue(callback, cachePropertyKeys, nodePath);
	}

	@NotNull
	@Override
	public String[] getNodeNames(@NotNull final String nodePath, @Nullable final Long nodeId) throws BackingStoreException {
		class CallbackNodeNames implements Callback<String[], BackingStoreException> {
			@Override
			public String[] call() throws BackingStoreException {
				return remotePreferencesService.getNodeNames(nodePath, nodeId);
			}
		}

		final CallbackNodeNames callback = new CallbackNodeNames();
		return getValue(callback, cacheNodeNames, nodePath);
	}

	@Nullable
	@Override
	public Long loadNode(@NotNull final String parentNodePath, @NotNull final Long parentNodeId, @NotNull final String childNodeName) {
		class CallbackNode implements Callback<Long, RuntimeException> {
			@Nullable
			@Override
			public Long call() {
				return remotePreferencesService.loadNode(parentNodePath, parentNodeId, childNodeName);
			}
		}

		final CallbackNode callback = new CallbackNode();
		return getValue(callback, cacheNode, parentNodePath + CACHE_KEY_SEPARATOR + childNodeName);
	}

	private <T, E extends Throwable> T getValue(Callback<T, E> callback, Cache cache, Object key) throws E {
		@Nullable final Cache.ValueWrapper valueWrapper = cache.get(key);
		if (valueWrapper != null) {
			return cast(valueWrapper.get());
		}

		final T loadedValue = callback.call();
		putValue(cache, key, loadedValue);

		return loadedValue;
	}

	private void putValue(Cache cache, Object key, Object value) {
		cache.put(key, value);
	}

	private void evictValue(Cache cache, Object key) {
		cache.evict(key);
	}

	private static interface Callback<T, E extends Throwable> {
		T call() throws E;
	}

	private static class EmptyCallback<T, E extends Throwable> implements Callback<T, E> {
		@Nullable
		@Override
		public T call() throws E {
			return null;
		}
	}

	@Override
	public void afterPropertiesSet() throws Exception {
		cacheRootNode = cacheManager.getCache(CACHE_NAME_ROOT_NODE);
		cacheNode = cacheManager.getCache(CACHE_NAME_NODE);
		cacheNodeNames = cacheManager.getCache(CACHE_NAME_NODE_NAMES);
		cacheProperty = cacheManager.getCache(CACHE_NAME_PROPERTY);
		cachePropertyKeys = cacheManager.getCache(CACHE_NAME_PROPERTY_KEYS);
	}

	public void setCacheManager(CacheManager cacheManager) {
		this.cacheManager = cacheManager;
	}

	public void setRemotePreferencesService(PreferencesService remotePreferencesService) {
		this.remotePreferencesService = remotePreferencesService;
	}
}
