package org.svaor.preferences;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.LinkedList;
import java.util.prefs.AbstractPreferences;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;

import static com.google.common.collect.Lists.newLinkedList;

/** @author Smelik Evgeniy */
public class PreferencesImpl extends AbstractPreferences {
	private static final Logger log = LoggerFactory.getLogger(PreferencesImpl.class);

	private static PreferencesImpl systemRoot;

	private Long nodeId;
	@Nullable
	private String username;

	//spring
	private static PreferencesService preferencesService;

	PreferencesImpl(AbstractPreferences parent, @NotNull String nodeName, @Nullable String username, @NotNull Long nodeId) {
		super(parent, nodeName);

		this.username = username;
		this.nodeId = nodeId;
	}

	PreferencesImpl(AbstractPreferences parent, @NotNull String nodeName, @Nullable String username) {
		super(parent, nodeName);

		this.username = username;
	}

	public static Preferences getUserRoot(String username) {
		if (log.isDebugEnabled()) {
			log.debug("Getting root preferences for user [" + username + ']');
		}

		final Long rootNodeId = preferencesService.loadOrCreateRootNode(username);

		return new PreferencesImpl(null, "", username, rootNodeId);
	}

	public static Preferences getSystemRoot() {
		log.debug("Getting preferences for system root");

		if (systemRoot == null) {
			log.debug("Creating preferences for system root");

			final Long rootNodeId = preferencesService.loadOrCreateRootNode(null);
			systemRoot = new PreferencesImpl(null, "", null, rootNodeId);
		}

		return systemRoot;
	}

	@Override
	protected void putSpi(@NotNull String key, @NotNull String value) {
		final LinkedList<String> nodes = newLinkedList();
		PreferencesImpl preferencesImpl = this;
		while (true) {
			nodes.addFirst(preferencesImpl.name());

			final PreferencesImpl parent = preferencesImpl.parent();
			if (parent.parent() == null) {
				break;
			} else {
				preferencesImpl = parent;
			}
		}

		final PreferencesImpl rootPreferences = preferencesImpl.parent();
		final Long rootNodeId = rootPreferences.getNodeId();
		assert rootNodeId != null;
		final Long[] ids = preferencesService.setProperty(rootNodeId, nodes.toArray(new String[nodes.size()]), absolutePath(), key, value);

		preferencesImpl = this;
		for (int i = ids.length - 1; i >= 0; i--) {
			preferencesImpl.setNodeId(ids[i]);
			preferencesImpl = preferencesImpl.parent();
		}
	}

	@Nullable
	@Override
	protected String getSpi(@NotNull String key) {
		return preferencesService.getProperty(absolutePath(), nodeId, key);
	}

	@Override
	protected void removeSpi(@NotNull String key) {
		preferencesService.removeProperty(absolutePath(), nodeId, key);
	}

	@Override
	protected void removeNodeSpi() throws BackingStoreException {
		preferencesService.removeNode(parent().absolutePath(), name(), nodeId);
	}

	@Override
	protected String[] keysSpi() throws BackingStoreException {
		return preferencesService.getPropertyKeys(absolutePath(), nodeId);
	}

	@Override
	protected String[] childrenNamesSpi() throws BackingStoreException {
		return preferencesService.getNodeNames(absolutePath(), nodeId);
	}

	@Override
	protected AbstractPreferences childSpi(@NotNull String name) {
		if (nodeId == null) {
			// Probably nodes have not been saved yet. We will do this later by saving appropriate property
			return new PreferencesImpl(this, name, username);
		} else {
			final Long childNodeId = preferencesService.loadNode(absolutePath(), nodeId, name);
			if (childNodeId == null) {
				return new PreferencesImpl(this, name, username);
			} else {
				return new PreferencesImpl(this, name, username, childNodeId);
			}
		}
	}

	@Override
	protected void syncSpi() throws BackingStoreException {
	}

	@Override
	protected void flushSpi() throws BackingStoreException {
	}

	@Override
	public PreferencesImpl node(String path) {
		return (PreferencesImpl) super.node(path);
	}

	@Override
	public PreferencesImpl parent() {
		return (PreferencesImpl) super.parent();
	}

	@Nullable
	public Long getNodeId() {
		return nodeId;
	}

	public void setNodeId(Long nodeId) {
		this.nodeId = nodeId;
	}

	public static void setPreferencesService(PreferencesService preferencesService) {
		PreferencesImpl.preferencesService = preferencesService;
	}
}
