package org.svaor.preferences.provider;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.ArrayUtils;
import org.hibernate.criterion.DetachedCriteria;
import org.hibernate.criterion.Restrictions;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.springframework.transaction.support.TransactionTemplate;
import org.springframework.util.Assert;
import org.svaor.preferences.PreferencesProvider;
import org.svaor.preferences.model.Node;
import org.svaor.preferences.model.Property;
import org.svaor.preferences.provider.model.NodeImpl;
import org.svaor.preferences.provider.model.PropertyImpl;

import java.io.Serializable;
import java.util.*;

import static com.google.common.collect.Lists.newLinkedList;
import static org.fluentjava.FluentUtils.cast;

/**
 * Preferences provider implementation based on hibernate ORM and EHCache
 * @author Smelik Evgeniy
 */
public class PreferencesProviderImpl implements PreferencesProvider {
	private final Logger log = LoggerFactory.getLogger(PreferencesProviderImpl.class);

	public static final String CACHE_USERNAME_ROOT_NODE_NAME = "preference-username-root-node";

	private static final int CACHE_USERNAME_ROOT_NODE_MAX_IN_MEMORY = 40960;
	private static final int CACHE_USERNAME_ROOT_NODE_TIME_TO_LIVE = 3600;
	private static final int CACHE_USERNAME_ROOT_NODE_TIME_TO_IDLE = 3600;

	private Cache cacheUsernameRootNode;
	private NullStub nullStub = new NullStub();

	//spring
	private HibernateTemplate hibernateTemplate;
	private TransactionTemplate transactionTemplate;
	private TransactionTemplate transactionTemplateReadOnly;

	public PreferencesProviderImpl() {
		try {
			final CacheManager cacheManager = CacheManager.getInstance();
			cacheUsernameRootNode = cacheManager.getCache(CACHE_USERNAME_ROOT_NODE_NAME);

			if (cacheUsernameRootNode == null) {
				log.warn("Creating default cache for preferences");

				cacheUsernameRootNode = new Cache(
						CACHE_USERNAME_ROOT_NODE_NAME, CACHE_USERNAME_ROOT_NODE_MAX_IN_MEMORY, false, true,
						CACHE_USERNAME_ROOT_NODE_TIME_TO_LIVE, CACHE_USERNAME_ROOT_NODE_TIME_TO_IDLE
				);
				cacheManager.addCache(cacheUsernameRootNode);
			}
		} catch (Exception e) {
			log.error(e.getMessage(), e);
		}
	}

	@NotNull
	@Override
	public Long loadOrCreateRootNode(final String username) {
		final Long cachedRootNodeId = getRootNodeIdFromCache(username);
		if (cachedRootNodeId != null) {
			return cachedRootNodeId;
		}

		final Long rootNodeId = transactionTemplate.execute(new TransactionCallback<Long>() {
			@Override
			public Long doInTransaction(TransactionStatus status) {
				final DetachedCriteria detachedCriteria = DetachedCriteria.forClass(NodeImpl.class);
				detachedCriteria.add(Restrictions.eq("username", username));
				detachedCriteria.add(Restrictions.isNull("parent"));

				//noinspection unchecked
				final List<Node> nodes = hibernateTemplate.findByCriteria(detachedCriteria);
				if (CollectionUtils.isNotEmpty(nodes)) {
					if (log.isDebugEnabled()) {
						log.debug("Loaded existent root node for username [" + username + ']');
					}

					return nodes.get(0).getId();
				}

				//create new node
				final NodeImpl newRootNode = new NodeImpl(NodeImpl.ROOT_NODE_NAME);
				newRootNode.setUsername(username);

				if (log.isDebugEnabled()) {
					log.debug("Creating new root node [" + newRootNode + ']');
				}

				return cast(hibernateTemplate.save(newRootNode));
			}
		});

		putRootNodeIdToCache(username, rootNodeId);

		return rootNodeId;
	}

	private static class NullStub implements Serializable {}

	@Nullable
	private Long getRootNodeIdFromCache(String username) {
		Assert.notNull(cacheUsernameRootNode);

		final Serializable key = username == null ? nullStub : username;
		try {
			final Element element = cacheUsernameRootNode.get(key);
			if (element == null) {
				if (log.isDebugEnabled()) {
					log.debug("Preferences cache miss: root node id is absent for username [" + username + ']');
				}

				return null;
			} else {
				final Long rootNodeId = cast(element.getValue());

				if (log.isDebugEnabled()) {
					log.debug("Preferences cache hit: root node id [" + rootNodeId + "] is present for username [" + username + ']');
				}

				return rootNodeId;
			}
		} catch (Exception e) {
			log.error("Error getting preferences root node id from cache for username [" + username + ']', e);
			return null;
		}
	}

	private void putRootNodeIdToCache(String username, Long rootNodeId) {
		Assert.notNull(cacheUsernameRootNode);

		final Serializable key = username == null ? nullStub : username;
		try {
			final Element element = cacheUsernameRootNode.get(key);
			if (element == null) {
				if (log.isDebugEnabled()) {
					log.debug("Adding root node id [" + rootNodeId + "] to cache for username [" + username + ']');
				}

				cacheUsernameRootNode.put(new Element(key, rootNodeId));
			}
		} catch (Exception e) {
			log.error("Error putting preferences root node id [" + rootNodeId + "] to cache for username [" + username + ']', e);
		}
	}

	@NotNull
	@Override
	public Long[] setProperty(@NotNull final Long rootNodeId, @Nullable final String[] nodeNames, @NotNull final String propertyName, @NotNull final String propertyValue) {
		return transactionTemplate.execute(new TransactionCallback<Long[]>() {
			@Override
			public Long[] doInTransaction(TransactionStatus status) {
				Node node = getNode(rootNodeId);
				if (ArrayUtils.isNotEmpty(nodeNames)) {
					assert nodeNames != null;
					for (String nodeName : nodeNames) {
						final Map<String, Node> childNodes = node.getChildNodes();
						final Node childNode = childNodes.get(nodeName);
						if (childNode == null) {
							node = new NodeImpl(nodeName, node);
						} else {
							node = childNode;
						}
					}
				}

				final LinkedList<Long> ids = newLinkedList();
				final Map<String, Property> properties = node.getProperties();

				for (Property property : properties.values()) {
					if (property.getName().equals(propertyName)) {
						property.setValue(propertyValue);
						property.setModifiedDate(new Date());

						if (log.isDebugEnabled()) {
							log.debug("Updating existing property [" + property + ']');
						}

						fillIds(ids, property);

						// property was found and updated
						return ids.toArray(new Long[ids.size()]);
					}
				}

				// property was not found, creating new one
				final PropertyImpl property = new PropertyImpl();
				property.setName(propertyName);
				property.setNode(node);
				property.setValue(propertyValue);
				properties.put(propertyName, property);

				if (log.isDebugEnabled()) {
					log.debug("Adding new property [" + property + ']');
				}

				hibernateTemplate.save(property);

				fillIds(ids, property);

				return ids.toArray(new Long[ids.size()]);
			}

			private void fillIds(LinkedList<Long> ids, Property property) {
				Node node = property.getNode();
				while (node != null) {
					ids.addFirst(node.getId());
					node = node.getParent();
				}
			}
		});
	}

	@Nullable
	@Override
	public String getProperty(@NotNull final Long nodeId, @NotNull final String propertyKey) {
		return transactionTemplateReadOnly.execute(new TransactionCallback<String>() {
			@Nullable
			@Override
			public String doInTransaction(TransactionStatus status) {
				final Node node = getNode(nodeId);
				final Map<String, Property> properties = node.getProperties();

				final Property property = properties.get(propertyKey);
				if (property == null) {
					if (log.isDebugEnabled()) {
						log.debug("Unknown property with key [" + propertyKey + "] for node [" + node + ']');
					}

					return null;
				}

				final String value = property.getValue();

				if (log.isDebugEnabled()) {
					log.debug("Loaded property with key [" + propertyKey + "] for node [" + node + "]. Value = [" + value + ']');
				}

				return value;
			}
		});
	}

	@Override
	public void removeProperty(@NotNull final Long nodeId, @NotNull final String propertyKey) {
		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node node = getNode(nodeId);
				final Map<String, Property> properties = node.getProperties();

				if (log.isDebugEnabled()) {
					log.debug("Deleting property with key [" + propertyKey + "] from node [" + node + ']');
				}

				properties.remove(propertyKey);

				if (MapUtils.isEmpty(node.getProperties()) && MapUtils.isEmpty(node.getChildNodes())) {
					removeNode(node.getId());
				}
			}
		});
	}

	@Override
	public void removeNode(@NotNull final Long nodeId) {
		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node node = getNode(nodeId);

				if (log.isDebugEnabled()) {
					log.debug("removeNode() is called for node [" + node + ']');
				}

				if (node.getParent() == null) {
					if (log.isDebugEnabled()) {
						log.debug("Skipping root node [" + node + "] removing");
					}

					return;
				}

				Node nodeForRemoving = node;

				Node parentNode = node.getParent();
				// удаляем все родительские узлы, которые остались без свойств и дочерних узлов, кроме корневого узла
				while (parentNode != null) {
					if (parentNode.getParent() != null && MapUtils.isEmpty(parentNode.getProperties()) && parentNode.getChildNodes().size() == 1) {
						if (log.isTraceEnabled()) {
							log.trace("Node [" + parentNode + "] left without children and will be deleted");
						}

						nodeForRemoving = parentNode;
						parentNode = parentNode.getParent();
					} else {
						parentNode = null;
					}
				}

				if (log.isDebugEnabled()) {
					log.debug("Deleting node [" + nodeForRemoving + ']');
				}

				final Node parent = nodeForRemoving.getParent();
				assert parent != null;
				final Map<String, Node> childNodes = parent.getChildNodes();
				childNodes.remove(nodeForRemoving.getName());
			}
		});
	}

	@NotNull
	@Override
	public String[] getPropertyKeys(@NotNull final Long nodeId) {
		return transactionTemplateReadOnly.execute(new TransactionCallback<String[]>() {
			@Override
			public String[] doInTransaction(TransactionStatus status) {
				final Node node = getNode(nodeId);
				final Set<String> propertyNames = node.getProperties().keySet();

				if (log.isDebugEnabled()) {
					log.debug("Loaded property names [" + propertyNames + "] for node [" + node + ']');
				}

				return propertyNames.toArray(new String[propertyNames.size()]);
			}
		});
	}

	@NotNull
	@Override
	public String[] getNodeNames(@NotNull final Long nodeId) {
		return transactionTemplateReadOnly.execute(new TransactionCallback<String[]>() {
			@Override
			public String[] doInTransaction(TransactionStatus status) {
				final Node node = getNode(nodeId);
				final Set<String> nodeNames = node.getChildNodes().keySet();

				final String[] nodeNamesArray = nodeNames.toArray(new String[nodeNames.size()]);

				if (log.isDebugEnabled()) {
					log.debug("Loaded node names [" + ArrayUtils.toString(nodeNamesArray) + "] for parent node [" + node + ']');
				}

				return nodeNamesArray;
			}
		});
	}

	@Nullable
	@Override
	public Long loadNode(@NotNull final Long parentNodeId, @NotNull final String childNodeName) {
		return transactionTemplateReadOnly.execute(new TransactionCallback<Long>() {
			@Nullable
			@Override
			public Long doInTransaction(TransactionStatus status) {
				final Node parentNode = getNode(parentNodeId);
				final Map<String, Node> childNodes = parentNode.getChildNodes();
				final Node existentNode = childNodes.get(childNodeName);
				if (existentNode == null) {
					return null;
				}

				if (log.isDebugEnabled()) {
					log.debug("Loaded existent node [" + existentNode + "] for parent node [" + parentNode + ']');
				}

				return existentNode.getId();
			}
		});
	}

	@NotNull
	private Node getNode(@NotNull final Long nodeId) {
		return hibernateTemplate.get(NodeImpl.class, nodeId);
	}

	public void setHibernateTemplate(HibernateTemplate hibernateTemplate) {
		this.hibernateTemplate = hibernateTemplate;
	}

	public void setTransactionTemplate(TransactionTemplate transactionTemplate) {
		this.transactionTemplate = transactionTemplate;
	}

	public void setTransactionTemplateReadOnly(TransactionTemplate transactionTemplateReadOnly) {
		this.transactionTemplateReadOnly = transactionTemplateReadOnly;
	}
}
