package org.svaor.preferences.utils;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.BeanCreationException;
import org.springframework.beans.factory.config.AbstractFactoryBean;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;

/**
 * This factory bean looks for required bean in existing application context and if it can not be found returns default bean instance.
 * 1. The required bean can be found by its template name. E.g. "transactionManager", "sessionFactory" ot "dataSource".
 * 2. The required bean can be found by its class. If application context contains single bean of this class current factory uses it.
 * If no one bean in current application context matches default instance is used.
 * @author Smelik Evgeniy
 */
public class OptionBeanFactoryBean<T> extends AbstractFactoryBean<T> implements ApplicationContextAware {
	@Nullable
	private String beanTypicalName;
	@NotNull
	private String defaultBeanName;
	@NotNull
	private Class<T> beanClass;
	private ApplicationContext applicationContext;

	@Override
	public Class<?> getObjectType() {
		return beanClass;
	}

	@Override
	protected T createInstance() throws Exception {
		T instance = searchSingleBeanByName();
		if (instance != null) {
			return instance;
		}

		instance = searchSingleBeanByClass();
		if (instance != null) {
			return instance;
		}

		instance = prepareDefaultBean();

		if (instance == null) {
			throw new BeanCreationException("Default bean with name [" + defaultBeanName + "] can not be created");
		}

		return instance;
	}

	@Nullable
	private T searchSingleBeanByName() {
		if (beanTypicalName == null) {
			return null;
		}

		return getBeanOfClass(beanTypicalName, beanClass);
	}

	@Nullable
	private T getBeanOfClass(String beanName, Class<T> beanClass) {
		if (beanName == null) {
			return null;
		}

		try {
			return applicationContext.getBean(beanName, beanClass);
		} catch (BeansException be) {
			return null;
		}
	}

	@Nullable
	private T searchSingleBeanByClass() {
		final String[] beanNames = applicationContext.getBeanNamesForType(beanClass, false, false);
		String requiredBeanName = null;
		int foundBeanCount = 0;
		for (String beanName : beanNames) {
			if (beanName.equals(defaultBeanName)) {
				continue;
			}

			foundBeanCount++;
			requiredBeanName = beanName;
		}

		if (foundBeanCount != 1) {
			return null;
		}

		return getBeanOfClass(requiredBeanName, beanClass);
	}

	private T prepareDefaultBean() {
		return applicationContext.getBean(defaultBeanName, beanClass);
	}

	@Override
	public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
		this.applicationContext = applicationContext;
	}

	public void setBeanTypicalName(@Nullable String beanTypicalName) {
		this.beanTypicalName = beanTypicalName;
	}

	public void setDefaultBeanName(@NotNull String defaultBeanName) {
		this.defaultBeanName = defaultBeanName;
	}

	public void setBeanClass(@NotNull Class<T> beanClass) {
		this.beanClass = beanClass;
	}
}
