package org.svaor.preferences;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.RandomStringUtils;
import org.apache.commons.lang.StringUtils;
import org.hibernate.criterion.DetachedCriteria;
import org.hibernate.criterion.Restrictions;
import org.jetbrains.annotations.Nullable;
import org.junit.Before;
import org.junit.Test;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.svaor.preferences.model.Node;
import org.svaor.preferences.model.Property;
import org.svaor.preferences.provider.model.NodeImpl;

import java.util.List;
import java.util.Map;
import java.util.Random;

import static junit.framework.Assert.*;
import static org.fluentjava.FluentUtils.cast;

/**
 * Abstract top-level test for database preferences API implementation.
 * @author Smelik Evgeniy
 */
public abstract class AbstractPreferencesTopLevelTest extends AbstractPreferencesTest {
	@Nullable
	protected String username;

	public AbstractPreferencesTopLevelTest() {
	}

	@Before
	public void before() {
		prepareUser();
		preparePreferences();
	}

	protected abstract void prepareUser();

	protected abstract void preparePreferences();

	protected Long prepareNodeId() {
		return transactionTemplate.execute(new TransactionCallback<Long>() {
			@Override
			public Long doInTransaction(TransactionStatus status) {
				final DetachedCriteria detachedCriteria = DetachedCriteria.forClass(NodeImpl.class);
				detachedCriteria.add(Restrictions.isNull("parent"));
				if (username == null) {
					detachedCriteria.add(Restrictions.isNull("username"));
				} else {
					detachedCriteria.add(Restrictions.eq("username", username));
				}

				final List nodes = hibernateTemplate.findByCriteria(detachedCriteria);
				final Node rootNode;
				if (CollectionUtils.isEmpty(nodes)) {
					rootNode = null;
				} else if (nodes.size() == 1) {
					rootNode = cast(nodes.get(0));
				} else {
					throw new IllegalStateException("More than one node found");
				}

				assertNotNull(rootNode);

				final String classPackage = CLASS_FOR_PREFERENCES.getPackage().getName();
				Node node = rootNode;
				for (String packageFragment : StringUtils.split(classPackage, '.')) {
					node = node.getChildNodes().get(packageFragment);
					assertNotNull("Node with name [" + packageFragment + "] does not exist", node);
					assertNotNull(node.getId());
				}

				return node.getId();
			}
		});
	}

	@Test
	public void testString() {
		final String testKeyName = "testKeyForString";
		final String stringValueDefault = "default", stringValueNewExpected = "nonDefault";
		final String stringValue = preferences.get(testKeyName, stringValueDefault);
		assertEquals(stringValueDefault, stringValue);

		preferences.put(testKeyName, stringValueNewExpected);
		final Long nodeId = prepareNodeId();
		checkPropertyExists(nodeId, testKeyName);

		String newStringValue = preferences.get(testKeyName, stringValue);
		assertEquals(stringValueNewExpected, newStringValue);
		newStringValue = preferences.get(testKeyName, stringValue);
		assertEquals(stringValueNewExpected, newStringValue);
	}

	@Test
	public void testLargeString() {
		final String testKeyName = "testKeyForLargeString";
		final String largeStringValueDefault = RandomStringUtils.randomAlphanumeric(4000), largeStringValueNewExpected = RandomStringUtils.randomAlphanumeric(4000);
		final String largeStringValue = preferences.get(testKeyName, largeStringValueDefault);
		assertEquals(largeStringValueDefault, largeStringValue);

		preferences.put(testKeyName, largeStringValueNewExpected);
		final Long nodeId = prepareNodeId();
		checkPropertyExists(nodeId, testKeyName);

		final String newLargeStringValue = preferences.get(testKeyName, largeStringValueDefault);
		assertEquals(largeStringValueNewExpected, newLargeStringValue);
	}

	@Test
	public void testBoolean() {
		final String testKeyName = "testKeyForBoolean";
		final boolean booleanValueDefault = true, booleanValueNewExpected = false;
		final boolean booleanValue = preferences.getBoolean(testKeyName, booleanValueDefault);
		assertEquals(booleanValueDefault, booleanValue);

		preferences.putBoolean(testKeyName, booleanValueNewExpected);
		final Long nodeId = prepareNodeId();
		checkPropertyExists(nodeId, testKeyName);

		final boolean newBooleanValue = preferences.getBoolean(testKeyName, booleanValueDefault);
		assertEquals(booleanValueNewExpected, newBooleanValue);
	}

	@Test
	public void testByteArray() {
		final Random generator = new Random();
		final String testKeyName = "testKeyForByteArray";
		final byte[] byteArrayValueDefault = new byte[200], byteArrayValueNewExpected = new byte[200];
		generator.nextBytes(byteArrayValueDefault);
		generator.nextBytes(byteArrayValueNewExpected);
		final byte[] byteArrayValue = preferences.getByteArray(testKeyName, byteArrayValueDefault);
		assertTrue(ArrayUtils.isEquals(byteArrayValueDefault, byteArrayValue));

		preferences.putByteArray(testKeyName, byteArrayValueNewExpected);
		final Long nodeId = prepareNodeId();
		checkPropertyExists(nodeId, testKeyName);

		final byte[] newByteArrayValue = preferences.getByteArray(testKeyName, byteArrayValueDefault);
		assertTrue(ArrayUtils.isEquals(byteArrayValueNewExpected, newByteArrayValue));
	}

	@Test
	public void testDouble() {
		final String testKeyName = "testKeyForDouble";
		final double doubleValueDefault = 1, doubleValueNewExpected = 2;
		final double doubleValue = preferences.getDouble(testKeyName, doubleValueDefault);
		assertEquals(doubleValueDefault, doubleValue);

		preferences.putDouble(testKeyName, doubleValueNewExpected);
		final Long nodeId = prepareNodeId();
		checkPropertyExists(nodeId, testKeyName);

		final double newDoubleValue = preferences.getDouble(testKeyName, doubleValueDefault);
		assertEquals(doubleValueNewExpected, newDoubleValue);
	}

	@Test
	public void testFloat() {
		final String testKeyName = "testKeyForFloat";
		final float floatValueDefault = 1, floatValueNewExpected = 2;
		final float floatValue = preferences.getFloat(testKeyName, floatValueDefault);
		assertEquals(floatValueDefault, floatValue);

		preferences.putFloat(testKeyName, floatValueNewExpected);
		final Long nodeId = prepareNodeId();
		checkPropertyExists(nodeId, testKeyName);

		final float newFloatValue = preferences.getFloat(testKeyName, floatValueDefault);
		assertEquals(floatValueNewExpected, newFloatValue);
	}

	@Test
	public void testInt() {
		final String testKeyName = "testKeyForInt";
		final int intValueDefault = 1, intValueNewExpected = 2;
		final int intValue = preferences.getInt(testKeyName, intValueDefault);
		assertEquals(intValueDefault, intValue);

		preferences.putInt(testKeyName, intValueNewExpected);
		final Long nodeId = prepareNodeId();
		checkPropertyExists(nodeId, testKeyName);

		final int newIntValue = preferences.getInt(testKeyName, intValueDefault);
		assertEquals(intValueNewExpected, newIntValue);
	}

	@Test
	public void testLong() {
		final String testKeyName = "testKeyForLong";
		final long longValueDefault = 1, longValueNewExpected = 2;
		final long longValue = preferences.getLong(testKeyName, longValueDefault);
		assertEquals(longValueDefault, longValue);

		preferences.putLong(testKeyName, longValueNewExpected);
		final Long nodeId = prepareNodeId();
		checkPropertyExists(nodeId, testKeyName);

		final long newLongValue = preferences.getLong(testKeyName, longValueDefault);
		assertEquals(longValueNewExpected, newLongValue);
	}

	private void checkPropertyExists(final Long nodeId, final String testKeyName) {
		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node node = hibernateTemplate.get(NodeImpl.class, nodeId);
				final Map<String, Property> properties = node.getProperties();
				assertNotNull(properties.get(testKeyName));
			}
		});
	}
}
