package org.svaor.preferences;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.hibernate.criterion.DetachedCriteria;
import org.hibernate.criterion.Restrictions;
import org.junit.Before;
import org.junit.Test;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.svaor.preferences.model.Node;
import org.svaor.preferences.provider.model.NodeImpl;

import java.util.List;
import java.util.prefs.Preferences;

import static junit.framework.Assert.*;
import static org.fluentjava.FluentUtils.cast;

/** @author Smelik Evgeniy */
public class PreferencesInitializationTest extends AbstractPreferencesTest {
	public PreferencesInitializationTest() {
	}

	@Before
	public void before() {
		preparePreferences();
	}

	@Test
	public void testNotCreatingEmptyNode() {
		final String testKeyName = "testKeyForUnsavedProperty";
		final String stringValueDefault = "default";
		final String stringValue = preferences.get(testKeyName, stringValueDefault);
		assertEquals(stringValueDefault, stringValue);

		// Nodes must not be generated while getting nonexistent property
		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final DetachedCriteria detachedCriteria = DetachedCriteria.forClass(NodeImpl.class);
				detachedCriteria.add(Restrictions.isNull("parent"));
				detachedCriteria.add(Restrictions.isNull("username"));

				final List nodes = hibernateTemplate.findByCriteria(detachedCriteria);
				final Node rootNode;
				if (CollectionUtils.isEmpty(nodes)) {
					rootNode = null;
				} else if (nodes.size() == 1) {
					rootNode = cast(nodes.get(0));
				} else {
					throw new IllegalStateException("More than one node found");
				}

				assertNotNull(rootNode);
				assertNotNull(rootNode.getChildNodes());
				assertTrue(MapUtils.isEmpty(rootNode.getChildNodes()));
			}
		});

		// PreferencesImpl must not have identifier yet
		PreferencesImpl curPreferences = cast(preferences);
		while (curPreferences != null) {
			final Preferences parent = curPreferences.parent();
			if (parent == null) {
				break;
			}
			assertNull("Identity = [" + curPreferences.getNodeId() + ']', curPreferences.getNodeId());
			curPreferences = cast(parent);
		}
	}

	@Test
	public void testCreatingNonemptyNode() {
		final String testKeyName = "testKeyForUnsavedProperty";
		final String stringValue = "nonDefault";

		// Saving property must store all nodes for it
		preferences.put(testKeyName, stringValue);

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final DetachedCriteria detachedCriteria = DetachedCriteria.forClass(NodeImpl.class);
				detachedCriteria.add(Restrictions.isNull("parent"));
				detachedCriteria.add(Restrictions.isNull("username"));

				final List nodes = hibernateTemplate.findByCriteria(detachedCriteria);
				final Node rootNode;
				if (CollectionUtils.isEmpty(nodes)) {
					rootNode = null;
				} else if (nodes.size() == 1) {
					rootNode = cast(nodes.get(0));
				} else {
					throw new IllegalStateException("More than one node found");
				}

				assertNotNull(rootNode);

				final String classPackage = CLASS_FOR_PREFERENCES.getPackage().getName();
				Node node = rootNode;
				for (String packageFragment : StringUtils.split(classPackage, '.')) {
					node = node.getChildNodes().get(packageFragment);
					assertNotNull("Node with name [" + packageFragment + "] does not exist", node);
					assertNotNull(node.getId());
				}
			}
		});

		// Node identifiers should be generated while saving their properties
		PreferencesImpl curPreferences = cast(preferences);
		while (curPreferences != null) {
			assertNotNull("Identity = [" + curPreferences.getNodeId() + ']', curPreferences.getNodeId());
			curPreferences = cast(curPreferences.parent());
		}
	}
}
