package org.svaor.preferences;

import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.criterion.Projections;
import org.hibernate.criterion.Restrictions;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.orm.hibernate3.HibernateCallback;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionTemplate;
import org.svaor.preferences.provider.model.NodeImpl;
import org.svaor.preferences.provider.model.PropertyImpl;

import java.sql.SQLException;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertTrue;

/**
 * Basic top-level preferences implementation test
 * @author Smelik Evgeniy
 */
public class PreferencesTest extends AbstractPreferencesLowLevelTest {
	@Autowired
	private HibernateTemplate hibernateTemplate;
	@Autowired
	@Qualifier("preferencesTransactionTemplate")
	private TransactionTemplate transactionTemplate;

	public PreferencesTest() {
	}

	@Test
	public void testRemoveNodeWithParents() throws BackingStoreException {
		final int[] idsCountBeforeCreating = getCounts();

		//java.util.prefs
		final PreferencesImpl preferences = (PreferencesImpl) Preferences.systemRoot().node("testRemoveNodeWithParents-thisOneShouldBeDeleted").node("removeNode").node("withParents");
		preferences.put("test", "test");

		final int[] idsCountAfterCreating = getCounts();
		assertEquals(1, idsCountAfterCreating[0] - idsCountBeforeCreating[0]);
		assertEquals(3, idsCountAfterCreating[1] - idsCountBeforeCreating[1]);

		preferences.removeNode();

		final int[] idsCountAfterRemoving = getCounts();
		assertTrue(idsCountAfterRemoving[0] == idsCountBeforeCreating[0]);
		assertTrue(idsCountAfterRemoving[1] == idsCountBeforeCreating[1]);
	}

	@Test
	public void testRemoveNode() throws BackingStoreException {
		final int[] idsCountBeforeCreating = getCounts();

		PreferencesImpl preferences = (PreferencesImpl) Preferences.systemRoot().node("testRemoveNode").node("removeNode");
		preferences.put("test", "test");
		preferences = (PreferencesImpl) Preferences.systemRoot().node("testRemoveNode").node("removeNode").node("thisOneShouldBeDeleted");
		preferences.put("test", "test");

		final int[] idsCountAfterCreating = getCounts();
		assertEquals(2, idsCountAfterCreating[0] - idsCountBeforeCreating[0]);
		assertEquals(3, idsCountAfterCreating[1] - idsCountBeforeCreating[1]);

		preferences.removeNode();

		final int[] idsCountAfterRemoving = getCounts();
		assertEquals(1, idsCountAfterCreating[0] - idsCountAfterRemoving[0]);
		assertEquals(1, idsCountAfterCreating[1] - idsCountAfterRemoving[1]);
	}

	@Test
	public void testRemoveNodeWithChildren() throws BackingStoreException {
		final int[] idsCountBeforeCreating = getCounts();

		final PreferencesImpl preferences1 = (PreferencesImpl) Preferences.systemRoot().node("testRemoveNodeWithChildren-thisOneShouldBeDeleted").node("removeNode");
		preferences1.put("test", "test");
		final PreferencesImpl preferences2 = (PreferencesImpl) Preferences.systemRoot().node("testRemoveNodeWithChildren-thisOneShouldBeDeleted").node("removeNode").node("withChildren");
		preferences2.put("test", "test");

		final int[] idsCountAfterCreating = getCounts();
		assertEquals(2, idsCountAfterCreating[0] - idsCountBeforeCreating[0]);
		assertEquals(3, idsCountAfterCreating[1] - idsCountBeforeCreating[1]);

		preferences1.removeNode();

		final int[] idsCountAfterRemoving = getCounts();
		assertEquals(2, idsCountAfterCreating[0] - idsCountAfterRemoving[0]);
		assertEquals(3, idsCountAfterCreating[1] - idsCountAfterRemoving[1]);
	}

	private int[] getCounts() {
		return transactionTemplate.execute(new TransactionCallback<int[]>() {
			@Override
			public int[] doInTransaction(TransactionStatus status) {
				return hibernateTemplate.execute(new HibernateCallback<int[]>() {
					@Override
					public int[] doInHibernate(Session session) throws HibernateException, SQLException {
						Criteria criteria = session.createCriteria(PropertyImpl.class);
						criteria.setProjection(Projections.count("id"));
						final Number id = (Number) criteria.uniqueResult();

						criteria = session.createCriteria(NodeImpl.class);
						criteria.setProjection(Projections.count("id"));
						criteria.add(Restrictions.isNotNull("parent"));

						return new int[]{id.intValue(), ((Number) criteria.uniqueResult()).intValue()};
					}
				});
			}
		});
	}
}
