package org.svaor.preferences.provider;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.mutable.MutableLong;
import org.hibernate.criterion.DetachedCriteria;
import org.hibernate.criterion.Restrictions;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.springframework.transaction.support.TransactionTemplate;
import org.svaor.preferences.AbstractPreferencesLowLevelTest;
import org.svaor.preferences.PreferencesProvider;
import org.svaor.preferences.model.Node;
import org.svaor.preferences.model.Property;
import org.svaor.preferences.provider.model.NodeImpl;
import org.svaor.preferences.provider.model.PropertyImpl;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import static junit.framework.Assert.*;
import static org.apache.commons.lang.RandomStringUtils.randomAlphabetic;

/** @author Smelik Evgeniy */
public class PreferencesProviderTest extends AbstractPreferencesLowLevelTest {
	private final Logger log = LoggerFactory.getLogger(PreferencesProviderTest.class);

	//spring
	@Autowired
	private PreferencesProvider preferencesProvider;
	@Autowired
	@Qualifier("preferencesHibernateTemplate")
	private HibernateTemplate hibernateTemplate;
	@Autowired
	@Qualifier("preferencesTransactionTemplate")
	private TransactionTemplate transactionTemplate;

	public PreferencesProviderTest() {
	}

	@Test
	public void testLoadOrCreateRootNode() throws Exception {
		final String testUsername = generateTestUsername();

		final DetachedCriteria detachedCriteria = DetachedCriteria.forClass(NodeImpl.class);
		detachedCriteria.add(Restrictions.eq("username", testUsername));
		detachedCriteria.add(Restrictions.isNull("parent"));

		// verify that there is not root
		final List rootNodes = hibernateTemplate.findByCriteria(detachedCriteria);
		assertTrue(CollectionUtils.isEmpty(rootNodes));

		// first time - root created
		final Long rootNodeId = preferencesProvider.loadOrCreateRootNode(testUsername);

		final Node newRootNode = hibernateTemplate.get(NodeImpl.class, rootNodeId);
		assertNotNull(newRootNode);
		assertEquals(testUsername, newRootNode.getUsername());

		// second time - existent root loaded
		final Long oldRootNodeId = preferencesProvider.loadOrCreateRootNode(testUsername);
		assertEquals(rootNodeId, oldRootNodeId);
	}

	@Test
	public void testSetProperty() throws Exception {
		final String testUsername = generateTestUsername();

		final Long rootNodeId = preferencesProvider.loadOrCreateRootNode(testUsername);
		assertNotNull(rootNodeId);

		final String propertyName = randomAlphabetic(10), propertyValue1 = randomAlphabetic(10), propertyValue2 = randomAlphabetic(10);
		if (log.isDebugEnabled()) {
			log.debug("Property name [" + propertyName + "], first property value [" + propertyValue1 + "], next property value [" + propertyValue2 + ']');
		}

		Long[] ids = preferencesProvider.setProperty(rootNodeId, null, propertyName, propertyValue1);
		assertNotNull(ids);
		if (log.isDebugEnabled()) {
			log.debug("IDs [" + ArrayUtils.toString(ids) + ']');
		}
		assertEquals(1, ids.length);
		for (Long id : ids) {
			assertNotNull(id);
		}

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node rootNode = hibernateTemplate.get(NodeImpl.class, rootNodeId);
				final Map<String, Property> properties = rootNode.getProperties();
				assertNotNull(properties);
				assertEquals(1, properties.size());

				final Property property = properties.get(propertyName);
				assertNotNull(property);
				assertEquals(propertyValue1, property.getValue());
			}
		});

		ids = preferencesProvider.setProperty(rootNodeId, null, propertyName, propertyValue2);
		assertNotNull(ids);
		if (log.isDebugEnabled()) {
			log.debug("IDs [" + ArrayUtils.toString(ids) + ']');
		}
		assertEquals(1, ids.length);
		for (Long id : ids) {
			assertNotNull(id);
		}

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node rootNode = hibernateTemplate.get(NodeImpl.class, rootNodeId);
				final Map<String, Property> properties = rootNode.getProperties();
				assertNotNull(properties);
				assertEquals(1, properties.size());

				final Property property = properties.get(propertyName);
				assertNotNull(property);
				assertEquals(propertyValue2, property.getValue());
			}
		});
	}

	@Test
	public void testGetProperty() throws Exception {
		final String testUsername = generateTestUsername();

		final Long rootNodeId = preferencesProvider.loadOrCreateRootNode(testUsername);
		assertNotNull(rootNodeId);

		final String propertyName = randomAlphabetic(10), propertyValue = randomAlphabetic(10);
		if (log.isDebugEnabled()) {
			log.debug("Property name [" + propertyName + "], property value [" + propertyValue + ']');
		}

		preferencesProvider.setProperty(rootNodeId, null, propertyName, propertyValue);

		final String providedPropertyValue = preferencesProvider.getProperty(rootNodeId, propertyName);
		assertEquals(propertyValue, providedPropertyValue);

		final String nonexistentPropertyValue = preferencesProvider.getProperty(rootNodeId, "nonexistent");
		assertNull(nonexistentPropertyValue);
	}

	@Test
	public void testRemoveProperty() throws Exception {
		final String testUsername = generateTestUsername();

		final Long rootNodeId = preferencesProvider.loadOrCreateRootNode(testUsername);
		assertNotNull(rootNodeId);

		final String propertyName = randomAlphabetic(10), propertyValue = randomAlphabetic(10);
		if (log.isDebugEnabled()) {
			log.debug("Property name [" + propertyName + "], property value [" + propertyValue + ']');
		}

		preferencesProvider.setProperty(rootNodeId, null, propertyName, propertyValue);

		final Long propertyId = transactionTemplate.execute(new TransactionCallback<Long>() {
			@Override
			public Long doInTransaction(TransactionStatus status) {
				final Node rootNode = hibernateTemplate.get(NodeImpl.class, rootNodeId);
				final Map<String, Property> properties = rootNode.getProperties();
				assertNotNull(properties);
				assertEquals(1, properties.size());

				final Property property = properties.get(propertyName);
				assertNotNull(property);
				assertNotNull(property.getId());

				return property.getId();
			}
		});

		preferencesProvider.removeProperty(rootNodeId, propertyName);

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node rootNode = hibernateTemplate.get(NodeImpl.class, rootNodeId);
				final Map<String, Property> properties = rootNode.getProperties();
				assertTrue(MapUtils.isEmpty(properties));

				final Property property = hibernateTemplate.get(PropertyImpl.class, propertyId);
				assertNull(property);
			}
		});
	}

	@Test
	public void testRemoveNode() throws Exception {
		final String testUsername = generateTestUsername();

		// verify that root node is not deleted. Deleting root nodes is senselessly
		final Long rootNodeId = preferencesProvider.loadOrCreateRootNode(testUsername);
		assertNotNull(rootNodeId);

		preferencesProvider.removeNode(rootNodeId);

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node rootNode = hibernateTemplate.get(NodeImpl.class, rootNodeId);
				assertNotNull(rootNode);
			}
		});

		// verify deleting simple node and cascade deleting all linked properties
		final String childNodeName = randomAlphabetic(10), propertyName = randomAlphabetic(10);
		if (log.isDebugEnabled()) {
			log.debug("Child node name [" + childNodeName + ']');
			log.debug("Property name [" + propertyName + ']');
		}

		preferencesProvider.setProperty(rootNodeId, new String[]{childNodeName}, propertyName, randomAlphabetic(10));

		final MutableLong propertyId = new MutableLong(), childNodeId = new MutableLong();
		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node rootNode = hibernateTemplate.get(NodeImpl.class, rootNodeId);
				assertNotNull(rootNode);

				final Map<String, Node> childNodes = rootNode.getChildNodes();
				final Node childNode = childNodes.get(childNodeName);
				assertNotNull(childNode);
				childNodeId.setValue(childNode.getId());

				final Map<String, Property> properties = childNode.getProperties();
				assertNotNull(properties);
				assertEquals(1, properties.size());

				final Property property = properties.get(propertyName);
				assertNotNull(property);
				assertNotNull(property.getId());

				propertyId.setValue(property.getId());
			}
		});

		preferencesProvider.removeNode(childNodeId.longValue());

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node childNode = hibernateTemplate.get(NodeImpl.class, childNodeId.longValue());
				assertNull(childNode);

				final Property property = hibernateTemplate.get(PropertyImpl.class, propertyId.longValue());
				assertNull(property);
			}
		});
	}

	@Test
	public void testGetPropertyKeys() throws Exception {
		final String testUsername = generateTestUsername();

		final Long rootNodeId = preferencesProvider.loadOrCreateRootNode(testUsername);
		assertNotNull(rootNodeId);

		// verify that array of property keys is empty in the absence of properties
		String[] propertyKeys = preferencesProvider.getPropertyKeys(rootNodeId);
		assertNotNull(propertyKeys);
		assertEquals(0, propertyKeys.length);

		// verify case with two properties
		final String propertyName1 = randomAlphabetic(10), propertyValue1 = randomAlphabetic(10);
		if (log.isDebugEnabled()) {
			log.debug("Property 1: name [" + propertyName1 + "], value [" + propertyValue1 + ']');
		}

		preferencesProvider.setProperty(rootNodeId, null, propertyName1, propertyValue1);

		final String propertyName2 = randomAlphabetic(10), propertyValue2 = randomAlphabetic(10);
		if (log.isDebugEnabled()) {
			log.debug("Property 1: name [" + propertyName2 + "], value [" + propertyValue2 + ']');
		}

		preferencesProvider.setProperty(rootNodeId, null, propertyName2, propertyValue2);

		propertyKeys = preferencesProvider.getPropertyKeys(rootNodeId);
		assertNotNull(propertyKeys);
		assertEquals(2, propertyKeys.length);
		assertTrue(ArrayUtils.contains(propertyKeys, propertyName1));
		assertTrue(ArrayUtils.contains(propertyKeys, propertyName2));
	}

	@Test
	public void testGetNodeNames() throws Exception {
		final String testUsername = generateTestUsername();

		final Long rootNodeId = preferencesProvider.loadOrCreateRootNode(testUsername);
		assertNotNull(rootNodeId);

		// verify that array of child node keys is empty in the absence of child nodes
		String[] nodeNames = preferencesProvider.getNodeNames(rootNodeId);
		assertNotNull(nodeNames);
		assertEquals(0, nodeNames.length);

		final String nodeName1 = randomAlphabetic(10), nodeName2 = randomAlphabetic(10);
		if (log.isDebugEnabled()) {
			log.debug("Node 1 name [" + nodeName1 + "] node 2 name [" + nodeName2 + ']');
		}

		preferencesProvider.setProperty(rootNodeId, new String[]{nodeName1}, randomAlphabetic(10), randomAlphabetic(10));
		preferencesProvider.setProperty(rootNodeId, new String[]{nodeName2}, randomAlphabetic(10), randomAlphabetic(10));

		nodeNames = preferencesProvider.getNodeNames(rootNodeId);
		assertNotNull(nodeNames);
		assertEquals(2, nodeNames.length);
		assertTrue(ArrayUtils.contains(nodeNames, nodeName1));
		assertTrue(ArrayUtils.contains(nodeNames, nodeName2));
	}

	@Test
	public void testCreateNode() throws Exception {
		final String testUsername = generateTestUsername();

		final Long rootNodeId = preferencesProvider.loadOrCreateRootNode(testUsername);
		assertNotNull(rootNodeId);

		final String nodeName1 = randomAlphabetic(10);
		if (log.isDebugEnabled()) {
			log.debug("Node 1 name [" + nodeName1 + ']');
		}

		Long childNodeId = preferencesProvider.loadNode(rootNodeId, nodeName1);
		assertNull(childNodeId);

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node rootNode = hibernateTemplate.get(NodeImpl.class, rootNodeId);
				assertNotNull(rootNode);

				final Node childNode = new NodeImpl(nodeName1, rootNode);
				hibernateTemplate.save(childNode);
			}
		});

		childNodeId = preferencesProvider.loadNode(rootNodeId, nodeName1);
		assertNotNull(childNodeId);

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Node rootNode = hibernateTemplate.get(NodeImpl.class, rootNodeId);
				assertNotNull(rootNode);

				final Map<String, Node> childNodes = rootNode.getChildNodes();
				assertNotNull(childNodes);
				assertEquals(1, childNodes.size());
				final Set<String> nodeNames = childNodes.keySet();
				assertTrue(nodeNames.contains(nodeName1));

				final Node childNode = childNodes.get(nodeName1);
				assertNotNull(childNode);
				assertEquals(testUsername, childNode.getUsername());
				assertEquals(nodeName1, childNode.getName());
				assertEquals(rootNode, childNode.getParent());
			}
		});
	}

	private String generateTestUsername() {
		final String testUsername = UUID.randomUUID().toString();
		if (log.isDebugEnabled()) {
			log.debug("Test username [" + testUsername + ']');
		}
		return testUsername;
	}
}
