package org.svaor.preferences.provider.model;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.RandomStringUtils;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.svaor.preferences.AbstractPreferencesLowLevelTest;
import org.svaor.preferences.model.Node;
import org.svaor.preferences.model.Property;

import static junit.framework.Assert.*;
import static org.fluentjava.FluentUtils.cast;

/**
 * Database preferences model test
 * @author Smelik Evgeniy
 */
public class PreferencesDataAccessTest extends AbstractPreferencesLowLevelTest {
	private final Logger log = LoggerFactory.getLogger(PreferencesDataAccessTest.class);

	private static final String NODE_NAME_ROOT = "node-name-test-root";
	private static final String NODE_NAME_CHILD1 = "node-name-test-child1";
	private static final String NODE_NAME_CHILD2 = "node-name-test-child2";
	private static final String NODE_NAME_CHILD3 = "node-name-test-child3";
	private static final String PROPERTY_NAME_ROOT1 = "property-name-test-root-property1";
	private static final String PROPERTY_VALUE_ROOT1 = "property-value-test-root-property1";
	private static final String PROPERTY_NAME_CHILD1_1 = "property-name-test-child1-property1";
	private static final String PROPERTY_VALUE_CHILD1_1 = "property-value-test-child1-property1";
	private static final String PROPERTY_NAME_CHILD1_2 = "property-name-test-child1-property2";
	private static final String PROPERTY_VALUE_CHILD1_2 = "property-value-test-child1-property2";

	public PreferencesDataAccessTest() {
	}

	@Test
	public void testAddingNewEmptyNode() {
		log.info("Storing new empty node [" + NODE_NAME_ROOT + ']');

		final String testUsername = RandomStringUtils.randomAlphabetic(10);

		final Long nodeRootId = transactionTemplate.execute(new TransactionCallback<Long>() {
			@Override
			public Long doInTransaction(TransactionStatus status) {
				final NodeImpl nodeRoot = new NodeImpl(NODE_NAME_ROOT);
				nodeRoot.setUsername(testUsername);

				return cast(hibernateTemplate.save(nodeRoot));
			}
		});

		if (log.isDebugEnabled()) {
			log.debug("Stored node with id [" + nodeRootId + ']');
		}

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final NodeImpl nodeRoot = hibernateTemplate.get(NodeImpl.class, nodeRootId);

				assertNotNull(nodeRoot);
				assertEquals(nodeRootId, (Long) nodeRoot.getId());
				assertEquals(testUsername, nodeRoot.getUsername());
				assertNull(nodeRoot.getParent());
				assertEquals(NODE_NAME_ROOT, nodeRoot.getName());
				assertTrue(MapUtils.isEmpty(nodeRoot.getChildNodes()));
				assertTrue(MapUtils.isEmpty(nodeRoot.getProperties()));
			}
		});
	}

	@Test
	public void testAddingNewNodeWithChildNodes() {
		log.info("Storing new node [" + NODE_NAME_ROOT + "] with child nodes");

		final String testUsername = RandomStringUtils.randomAlphabetic(10);

		final Long rootNodeId = transactionTemplate.execute(new TransactionCallback<Long>() {
			@Override
			public Long doInTransaction(TransactionStatus status) {
				final NodeImpl nodeRoot = new NodeImpl(NODE_NAME_ROOT);
				nodeRoot.setUsername(testUsername);

				final Long nodeRootId = cast(hibernateTemplate.save(nodeRoot));

				final NodeImpl nodeChild1 = new NodeImpl(NODE_NAME_CHILD1, nodeRoot);
				hibernateTemplate.save(nodeChild1);

				final NodeImpl nodeChild2 = new NodeImpl(NODE_NAME_CHILD2, nodeRoot);
				hibernateTemplate.save(nodeChild2);

				final NodeImpl nodeChild3 = new NodeImpl(NODE_NAME_CHILD3, nodeChild2);
				hibernateTemplate.save(nodeChild3);

				return nodeRootId;
			}
		});

		if (log.isDebugEnabled()) {
			log.debug("Stored node with id [" + rootNodeId + ']');
		}

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final NodeImpl nodeRoot = hibernateTemplate.get(NodeImpl.class, rootNodeId);

				assertNotNull(nodeRoot);
				assertEquals(rootNodeId, (Long) nodeRoot.getId());
				assertEquals(testUsername, nodeRoot.getUsername());
				assertNull(nodeRoot.getParent());
				assertEquals(NODE_NAME_ROOT, nodeRoot.getName());
				assertTrue(MapUtils.isNotEmpty(nodeRoot.getChildNodes()));
				assertEquals(2, nodeRoot.getChildNodes().size());
				assertTrue(MapUtils.isEmpty(nodeRoot.getProperties()));

				final Node nodeChild1 = nodeRoot.getChildNodes().get(NODE_NAME_CHILD1);
				assertNotNull(nodeChild1);
				assertEquals(testUsername, nodeChild1.getUsername());
				assertNotNull(nodeChild1.getParent());
				assertEquals(nodeRoot, nodeChild1.getParent());
				assertEquals(NODE_NAME_CHILD1, nodeChild1.getName());
				assertTrue(MapUtils.isEmpty(nodeChild1.getChildNodes()));
				assertTrue(MapUtils.isEmpty(nodeChild1.getProperties()));

				final Node nodeChild2 = nodeRoot.getChildNodes().get(NODE_NAME_CHILD2);
				assertNotNull(nodeChild2);
				assertNotNull(nodeChild2.getParent());
				assertEquals(nodeRoot, nodeChild2.getParent());
				assertEquals(NODE_NAME_CHILD2, nodeChild2.getName());
				assertTrue(MapUtils.isNotEmpty(nodeChild2.getChildNodes()));
				assertEquals(1, nodeChild2.getChildNodes().size());
				assertTrue(MapUtils.isEmpty(nodeChild2.getProperties()));

				final Node nodeChild3 = nodeChild2.getChildNodes().get(NODE_NAME_CHILD3);
				assertNotNull(nodeChild3);
				assertNotNull(nodeChild3.getParent());
				assertEquals(nodeChild2, nodeChild3.getParent());
				assertEquals(NODE_NAME_CHILD3, nodeChild3.getName());
				assertTrue(MapUtils.isEmpty(nodeChild3.getChildNodes()));
				assertTrue(MapUtils.isEmpty(nodeChild3.getProperties()));
			}
		});
	}

	@Test
	public void testAddingNewNodeWithChildNodesAndProperties() {
		log.info("Storing new node [" + NODE_NAME_ROOT + "] with child nodes and properties");
		final Long nodeRootId = transactionTemplate.execute(new TransactionCallback<Long>() {
			@Override
			public Long doInTransaction(TransactionStatus status) {
				final NodeImpl nodeRoot = new NodeImpl(NODE_NAME_ROOT);

				final Long nodeRootId = cast(hibernateTemplate.save(nodeRoot));

				final NodeImpl nodeChild1 = new NodeImpl(NODE_NAME_CHILD1, nodeRoot);
				hibernateTemplate.save(nodeChild1);

				final NodeImpl nodeChild2 = new NodeImpl(NODE_NAME_CHILD2, nodeRoot);
				hibernateTemplate.save(nodeChild2);

				PropertyImpl property = new PropertyImpl(PROPERTY_NAME_ROOT1, PROPERTY_VALUE_ROOT1, nodeRoot);
				hibernateTemplate.save(property);

				property = new PropertyImpl(PROPERTY_NAME_CHILD1_1, PROPERTY_VALUE_CHILD1_1, nodeChild1);
				hibernateTemplate.save(property);

				property = new PropertyImpl(PROPERTY_NAME_CHILD1_2, PROPERTY_VALUE_CHILD1_2, nodeChild1);
				hibernateTemplate.save(property);

				return nodeRootId;
			}
		});

		if (log.isDebugEnabled()) {
			log.debug("Stored node with id [" + nodeRootId + ']');
		}

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final NodeImpl nodeRoot = hibernateTemplate.get(NodeImpl.class, nodeRootId);

				assertNotNull(nodeRoot);
				assertTrue(MapUtils.isNotEmpty(nodeRoot.getProperties()));
				assertEquals(1, nodeRoot.getProperties().size());

				Property property = nodeRoot.getProperties().get(PROPERTY_NAME_ROOT1);
				assertNotNull(property);
				assertEquals(PROPERTY_NAME_ROOT1, property.getName());
				assertEquals(PROPERTY_VALUE_ROOT1, property.getValue());

				final Node nodeChild1 = nodeRoot.getChildNodes().get(NODE_NAME_CHILD1);
				assertNotNull(nodeChild1);
				assertTrue(MapUtils.isNotEmpty(nodeChild1.getProperties()));
				assertEquals(2, nodeChild1.getProperties().size());

				property = nodeChild1.getProperties().get(PROPERTY_NAME_CHILD1_1);
				assertNotNull(property);
				assertEquals(PROPERTY_NAME_CHILD1_1, property.getName());
				assertEquals(PROPERTY_VALUE_CHILD1_1, property.getValue());

				property = nodeChild1.getProperties().get(PROPERTY_NAME_CHILD1_2);
				assertNotNull(property);
				assertEquals(PROPERTY_NAME_CHILD1_2, property.getName());
				assertEquals(PROPERTY_VALUE_CHILD1_2, property.getValue());
			}
		});
	}

	@Test
	public void testAddingNewProperty() {
		log.info("Storing new property [" + PROPERTY_NAME_CHILD1_1 + ']');
		final Long propertyId = transactionTemplate.execute(new TransactionCallback<Long>() {
			@Override
			public Long doInTransaction(TransactionStatus status) {
				final NodeImpl nodeRoot = new NodeImpl(NODE_NAME_ROOT);
				final NodeImpl nodeChild1 = new NodeImpl(NODE_NAME_CHILD1, nodeRoot);
				final NodeImpl nodeChild2 = new NodeImpl(NODE_NAME_CHILD2, nodeChild1);

				final PropertyImpl property = new PropertyImpl(PROPERTY_NAME_CHILD1_1, PROPERTY_VALUE_CHILD1_1, nodeChild2);

				return cast(hibernateTemplate.save(property));
			}
		});

		if (log.isDebugEnabled()) {
			log.debug("Stored property with id [" + propertyId + ']');
		}

		transactionTemplate.execute(new TransactionCallbackWithoutResult() {
			@Override
			protected void doInTransactionWithoutResult(TransactionStatus status) {
				final Property property = hibernateTemplate.get(PropertyImpl.class, propertyId);

				assertNotNull(property);
				assertEquals(property.getName(), PROPERTY_NAME_CHILD1_1);
				assertEquals(property.getValue(), PROPERTY_VALUE_CHILD1_1);

				Node node = property.getNode();
				assertNotNull(node);
				assertEquals(node.getName(), NODE_NAME_CHILD2);

				node = node.getParent();
				assertNotNull(node);
				assertEquals(node.getName(), NODE_NAME_CHILD1);

				node = node.getParent();
				assertNotNull(node);
				assertEquals(node.getName(), NODE_NAME_ROOT);
				assertNull(node.getParent());
			}
		});
	}
}
